#![allow(unused)]

use anyhow::Result;
use std::marker::PhantomData;
use std::{ops::Deref, ops::DerefMut};

use crate::*;

#[derive(Copy, Clone)]
pub struct Query<C, F = EmptyFilter>
where
    F: StaticCheckFilter,
{
    entity: Entity,
    c: C,
    filter: PhantomData<F>,
}

impl<C, F: StaticCheckFilter> Query<C, F> {
    pub fn entity(&self) -> Entity {
        self.entity
    }
}

impl<C, F: StaticCheckFilter> Deref for Query<C, F> {
    type Target = C;

    fn deref(&self) -> &Self::Target {
        &self.c
    }
}

impl<C, F: StaticCheckFilter> DerefMut for Query<C, F> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.c
    }
}

trait CreateQueryArchetype<T, R, Func, Filter> {
    fn create(f: Func) -> Self;
}

pub trait StaticCheckFilter: Send + Sync + Default + Clone {
    fn check(entity: &EntityObject) -> bool;
    fn verify_dedup<O: EntityComponent>();
}

impl StaticCheckFilter for EmptyFilter {
    fn check(_entity: &EntityObject) -> bool {
        true
    }

    fn verify_dedup<O>() {}
}

pub struct QueryArchetype {
    check_entity: Box<dyn Fn(&EntityObject) -> bool + Send + Sync>,
    create_callback: Box<
        dyn Fn(
                &EntityObject,
            )
                -> Result<Box<dyn Fn(&mut Commands, &mut World) -> Result<()> + Send + Sync>>
            + Send
            + Sync,
    >,

    entities: Vec<Box<dyn Fn(&mut Commands, &mut World) -> Result<()> + Send + Sync>>,
}

macro_rules! impl_query_singleton_update {
    ( $($var:ident $(,)?)+ ) => {
        impl<Func, Filter, $( $var, )+> CreateQueryArchetype<( $( $var, )+ ), (), Func, Filter> for QueryArchetype
        where
            Func: Fn(&mut Commands, Query<( $( &mut $var, )+ ), Filter>) -> Result<()> + Send + Sync + Clone + 'static,
            Filter: StaticCheckFilter,
            $(
                $var: EntityComponent + ComponentDebug,
            )+
        {
            paste::item! {
                fn create(f: Func) -> Self {
                    $(
                        Filter::verify_dedup::<$var>();
                    )+

                    Self {
                        check_entity: Box::new({
                            move |entity| {
                                $(
                                    if !entity.components.contains::<$var>() {
                                        return false;
                                    }
                                )+

                                if !Filter::check(entity) {
                                    return false;
                                }

                                true
                            }
                        }),

                        create_callback: Box::new(move |entity| {
                            $(
                                let [< $var:lower >] = UnsafeComponentStore::from(
                                    entity.get_component::<$var>()?
                                );
                            )+

                            let e = entity.as_entity();
                            let f = f.clone();

                            Ok(Box::new(move |commands, _world| {
                                let q = Query {
                                    entity: e,
                                    c:  unsafe { ($([< $var:lower >].as_mut(),)+) },
                                    filter: PhantomData,
                                };

                                f(commands, q)
                            }))
                        }),

                        entities: Vec::new(),
                    }
                }
            }
        }
    };
}

impl_query_singleton_update!(R, U);

struct T {}

impl EntityComponent for T {
    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for T {
    fn debug_name() -> &'static str {
        "T"
    }
}

struct U {}

impl EntityComponent for U {
    fn name(&self) -> &str {
        Self::debug_name()
    }
}

impl ComponentDebug for U {
    fn debug_name() -> &'static str {
        "U"
    }
}

fn system(_commands: &mut Commands, mut _query: Query<(&mut T, &mut U)>) -> Result<()> {
    Ok(())
}

fn test() {
    let q = QueryArchetype::create(system);
}
